#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for full Water Resource Recovery Facility with ASM2d and ADM1
(WRRF; a.k.a., wastewater treatment plant) flowsheet example with ASM1 and ADM1.
The flowsheet follows the same formulation as benchmark simulation model no.2 (BSM2)
but comprises different specifications for default values than BSM2.
"""

# Some more information about this module
__author__ = "Chenyu Wang"

import platform
import pytest

from pyomo.environ import value

from idaes.core.util import DiagnosticsToolbox
from idaes.core.util.scaling import (
    get_jacobian,
    jacobian_cond,
)

from watertap.flowsheets.full_water_resource_recovery_facility.BSM2_P_extension import (
    main,
)
from watertap.core.solvers import get_solver

solver = get_solver()

is_reference_platform = (
    platform.system() == "Windows" and platform.python_version_tuple()[0] == "3"
)
is_linux_platform = (
    platform.system() == "Linux" and platform.python_version_tuple()[0] == "3"
)

reference_platform_only = pytest.mark.xfail(
    condition=(not is_reference_platform),
    run=True,
    strict=False,
    reason="These tests are expected to pass only on the reference platform (Python 3 on Windows)",
)

linux_platform_only = pytest.mark.xfail(
    condition=(not is_linux_platform),
    run=True,
    strict=False,
    reason="These tests are expected to pass only on the Linux platform (Python 3)",
)


@pytest.mark.requires_idaes_solver
class TestFullFlowsheetBioPFalse:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m, res, sm = main(bio_P=False)
        return m

    @pytest.mark.component
    def test_structural_issues(self, system_frame):
        dt = DiagnosticsToolbox(system_frame)
        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.solver
    @pytest.mark.component
    def test_numerical_issues(self, system_frame):
        dt = DiagnosticsToolbox(system_frame)
        warnings, _ = dt._collect_numerical_warnings()
        assert len(warnings) == 1
        assert "WARNING: 3 Variables at or outside bounds (tol=0.0E+00)" in warnings

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame

        assert value(m.fs.Treated.properties[0].flow_vol) == pytest.approx(
            0.24219, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_A"]) == pytest.approx(
            2.7392e-06, abs=1e-6
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_F"]) == pytest.approx(
            0.00027924, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_I"]) == pytest.approx(
            0.057450006, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_N2"]
        ) == pytest.approx(0.070789, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NH4"]
        ) == pytest.approx(0.00014979, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NO3"]
        ) == pytest.approx(0.008443, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_O2"]
        ) == pytest.approx(0.0011858, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_PO4"]
        ) == pytest.approx(0.703159, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_K"]) == pytest.approx(
            0.367442, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_Mg"]
        ) == pytest.approx(0.018463, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_IC"]
        ) == pytest.approx(0.149736, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_AUT"]
        ) == pytest.approx(0.00074563, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_H"]) == pytest.approx(
            0.0139618, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_I"]) == pytest.approx(
            0.0122129, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PAO"]
        ) == pytest.approx(0.0120675, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PHA"]
        ) == pytest.approx(5.5524e-06, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PP"]
        ) == pytest.approx(0.00402767, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_S"]) == pytest.approx(
            0.00022285, rel=1e-3
        )

        # Check electricity consumption for each aerobic reactor
        assert value(m.fs.R5.electricity_consumption[0]) == pytest.approx(
            133.333, rel=1e-3
        )
        assert value(m.fs.R6.electricity_consumption[0]) == pytest.approx(
            133.333, rel=1e-3
        )
        assert value(m.fs.R7.electricity_consumption[0]) == pytest.approx(
            46.666, rel=1e-3
        )

    @pytest.mark.component
    def test_costing(self, system_frame):
        m = system_frame

        # check costing
        assert value(m.fs.costing.LCOW) == pytest.approx(0.483273, rel=1e-3)
        assert value(m.fs.costing.total_capital_cost) == pytest.approx(
            24026877.393, rel=1e-3
        )
        assert value(m.fs.costing.total_operating_cost) == pytest.approx(
            923153.285, rel=1e-3
        )

    @pytest.mark.solver
    @pytest.mark.component
    def test_condition_number(self, system_frame):
        m = system_frame

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(m, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            1.0671e21, rel=1e-3
        )


@pytest.mark.requires_idaes_solver
class TestFullFlowsheetBioPTrue:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m, res, sm = main(bio_P=True)
        return m

    @pytest.mark.component
    def test_structural_issues(self, system_frame):
        dt = DiagnosticsToolbox(system_frame)
        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.solver
    @pytest.mark.component
    def test_numerical_issues(self, system_frame):
        dt = DiagnosticsToolbox(system_frame)
        warnings, _ = dt._collect_numerical_warnings()
        assert len(warnings) == 1
        assert "WARNING: 3 Variables at or outside bounds (tol=0.0E+00)" in warnings

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame

        assert value(m.fs.Treated.properties[0].flow_vol) == pytest.approx(
            0.2422, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_A"]) == pytest.approx(
            2.8676e-06, abs=1e-6
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_F"]) == pytest.approx(
            0.0002783, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_I"]) == pytest.approx(
            0.057450, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_N2"]
        ) == pytest.approx(0.0707506, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NH4"]
        ) == pytest.approx(0.0001517, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NO3"]
        ) == pytest.approx(0.0086813, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_O2"]
        ) == pytest.approx(0.0011696, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_PO4"]
        ) == pytest.approx(0.002821278, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_K"]) == pytest.approx(
            0.37, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_Mg"]
        ) == pytest.approx(0.0208324, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_IC"]
        ) == pytest.approx(0.14592, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_AUT"]
        ) == pytest.approx(0.0007451, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_H"]) == pytest.approx(
            0.01392289, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_I"]) == pytest.approx(
            0.0121976, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PAO"]
        ) == pytest.approx(0.0118308, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PHA"]
        ) == pytest.approx(5.70036e-06, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PP"]
        ) == pytest.approx(0.0038949, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_S"]) == pytest.approx(
            0.00022138, rel=1e-3
        )

        # Check electricity consumption for each aerobic reactor
        assert value(m.fs.R5.electricity_consumption[0]) == pytest.approx(
            133.3333, rel=1e-3
        )
        assert value(m.fs.R6.electricity_consumption[0]) == pytest.approx(
            133.333, rel=1e-3
        )
        assert value(m.fs.R6.electricity_consumption[0]) == pytest.approx(
            133.333, rel=1e-3
        )

    @pytest.mark.component
    def test_costing(self, system_frame):
        m = system_frame

        # check costing
        assert value(m.fs.costing.LCOW) == pytest.approx(0.4828, rel=1e-3)
        assert value(m.fs.costing.total_capital_cost) == pytest.approx(
            24003751.225, rel=1e-3
        )
        assert value(m.fs.costing.total_operating_cost) == pytest.approx(
            922338.478, rel=1e-3
        )

    @pytest.mark.solver
    @pytest.mark.component
    @reference_platform_only
    def test_condition_number(self, system_frame):
        m = system_frame

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(m, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            2.427164e21, rel=1e-2
        )


@pytest.mark.requires_idaes_solver
class TestScaledBioPFalse:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m, res, sm = main(bio_P=False)
        return sm

    @pytest.mark.component
    def test_structural_issues(self, system_frame):
        dt = DiagnosticsToolbox(system_frame)
        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.solver
    @pytest.mark.component
    def test_numerical_issues(self, system_frame):
        dt = DiagnosticsToolbox(system_frame)
        warnings, next_steps = dt._collect_numerical_warnings()
        assert len(warnings) == 1
        assert "WARNING: 3 Variables at or outside bounds (tol=0.0E+00)" in warnings

    @pytest.mark.solver
    @pytest.mark.component
    @linux_platform_only
    def test_condition_number_on_linux(self, system_frame):
        m = system_frame

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(m, scaled=False)
        assert jacobian_cond(jac=jac, scaled=False) == pytest.approx(
            2.987650e15, rel=1e-2
        )

    @pytest.mark.solver
    @pytest.mark.component
    @reference_platform_only
    def test_condition_number_on_windows(self, system_frame):
        m = system_frame

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(m, scaled=False)
        cond = jacobian_cond(jac=jac, scaled=False)
        assert (
            # Python 3.9-3.11
            cond == pytest.approx(4.3021828e15, rel=1e-2)
            # Python 3.12
            or cond == pytest.approx(2.987651e15, rel=1e-2)
        )


@pytest.mark.requires_idaes_solver
class TestScaledBioPTrue:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m, res, sm = main(bio_P=True)
        return sm

    @pytest.mark.component
    def test_structural_issues(self, system_frame):
        dt = DiagnosticsToolbox(system_frame)
        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.solver
    @pytest.mark.component
    def test_numerical_issues(self, system_frame):
        sm = system_frame
        dt = DiagnosticsToolbox(sm)
        warnings, next_steps = dt._collect_numerical_warnings()

        assert len(warnings) == 1
        assert "WARNING: 3 Variables at or outside bounds (tol=0.0E+00)" in warnings

    @pytest.mark.solver
    @pytest.mark.component
    @linux_platform_only
    def test_condition_number_on_linux(self, system_frame):
        sm = system_frame
        dt = DiagnosticsToolbox(sm)

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(sm, scaled=False)

        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            7.42017e14, rel=1e-3
        )

    @pytest.mark.solver
    @pytest.mark.component
    @reference_platform_only
    def test_condition_number_on_windows(self, system_frame):
        sm = system_frame
        dt = DiagnosticsToolbox(sm)

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(sm, scaled=False)

        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            7.43640e14, rel=1e-2
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a centrifuge unit.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Marcus Holly"


@declare_process_block_class("CentrifugeZO")
class CentrifugeZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a centrifuge reactor unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "centrifuge"

        build_sido(self)
        constant_intensity(self)

        self.polymer_dose = Var(
            self.flowsheet().time,
            units=pyunits.mg / pyunits.L,
            bounds=(0, None),
            doc="Dosing rate of polymer",
        )
        self._fixed_perf_vars.append(self.polymer_dose)
        self._perf_var_dict["Dosage of polymer per sludge"] = self.polymer_dose

        self.polymer_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Consumption rate of polymer",
        )
        self._perf_var_dict["Polymer Demand"] = self.polymer_demand

        @self.Constraint(self.flowsheet().time, doc="Polymer demand constraint")
        def polymer_demand_equation(b, t):
            return b.polymer_demand[t] == pyunits.convert(
                b.polymer_dose[t] * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

    @property
    def default_costing_method(self):
        return self.cost_centrifuge

    @staticmethod
    def cost_centrifuge(blk):
        """
        Method for costing centrifuge unit.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        HRT, size_cost = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["HRT", "sizing_cost"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol * HRT * size_cost,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.polymer_demand[t0], "polymer"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Inherits from a modified CSTR model with injection terms. This model assumes oxygen will be injected.
"""

# Import Pyomo libraries
from pyomo.common.config import In
from pyomo.environ import Constraint

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
)
from idaes.core.scaling import CustomScalerBase, ConstraintScalingScheme


from watertap.unit_models.cstr_injection import (
    CSTR_InjectionData,
    ElectricityConsumption,
)

__author__ = "Adam Atia"


class AerationTankScaler(CustomScalerBase):
    """
    Default modular scaler for the aeration tank unit model.

    This Scaler relies on the associated property and reaction packages,
    either through user provided options (submodel_scalers argument) or by default
    Scalers assigned to the packages.
    """

    DEFAULT_SCALING_FACTORS = {
        "volume": 1e-3,
        "hydraulic_retention_time": 1e-3,
        "KLa": 1e-1,
        "mass_transfer_term": 1e2,
    }

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        """
        Routine to apply scaling factors to variables in model.

        Args:
            model: model to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: dict of Scalers to use for sub-models, keyed by submodel local name

        Returns:
            None
        """
        # Call scaling methods for sub-models
        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_in,
            method="variable_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )
        self.propagate_state_scaling(
            target_state=model.control_volume.properties_out,
            source_state=model.control_volume.properties_in,
            overwrite=overwrite,
        )

        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_out,
            method="variable_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )
        self.call_submodel_scaler_method(
            submodel=model.control_volume.reactions,
            method="variable_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )

        # Scaling control volume variables
        self.scale_variable_by_default(
            model.control_volume.volume[0], overwrite=overwrite
        )
        self.scale_variable_by_default(
            model.hydraulic_retention_time[0], overwrite=overwrite
        )
        self.scale_variable_by_default(model.KLa, overwrite=overwrite)
        if model.config.has_aeration:
            if "S_O" in model.config.property_package.component_list:
                self.scale_variable_by_default(
                    model.control_volume.mass_transfer_term[0, "Liq", "S_O"],
                    overwrite=overwrite,
                )
            elif "S_O2" in model.config.property_package.component_list:
                self.scale_variable_by_default(
                    model.control_volume.mass_transfer_term[0, "Liq", "S_O2"],
                    overwrite=overwrite,
                )
            else:
                pass

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        """
        Routine to apply scaling factors to constraints in model.

        Submodel Scalers are called for the property and reaction blocks. All other constraints
        are scaled using the inverse maximum scheme.

        Args:
            model: model to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: dict of Scalers to use for sub-models, keyed by submodel local name

        Returns:
            None
        """
        # Call scaling methods for sub-models
        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_in,
            method="constraint_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )
        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_out,
            method="constraint_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )
        self.call_submodel_scaler_method(
            submodel=model.control_volume.reactions,
            method="constraint_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )

        # Scale unit level constraints
        for c in model.component_data_objects(Constraint, descend_into=True):
            self.scale_constraint_by_nominal_value(
                c,
                scheme=ConstraintScalingScheme.inverseMaximum,
                overwrite=overwrite,
            )


@declare_process_block_class("AerationTank")
class AerationTankData(CSTR_InjectionData):
    """
    CSTR Unit Model with Injection Class
    """

    default_scaler = AerationTankScaler

    CONFIG = CSTR_InjectionData.CONFIG()
    CONFIG.electricity_consumption = ElectricityConsumption.calculated
    CONFIG.get("has_aeration")._domain = In([True])
    CONFIG.has_aeration = True



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
class FrozenPipes(Exception):
    """
    Custom exception for WaterTAP developer errors.

    Raised when there's an issue that indicates a problem with the model
    implementation or configuration, rather than user input.

    Parameters
    ----------
    message : str
        Error message describing the issue
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"FrozenPipes: {self.message}"



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import re
import pytest

from pyomo.environ import (
    assert_optimal_termination,
    ConcreteModel,
    Constraint,
    Set,
    Suffix,
    TransformationFactory,
    value,
    Var,
)
from pyomo.network import Port

from idaes.core import EnergyBalanceType, FlowsheetBlock, MomentumBalanceType
from idaes.core.initialization import InitializationStatus
from idaes.core.scaling import set_scaling_factor
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.model_diagnostics import DiagnosticsToolbox
from idaes.core.util.testing import PhysicalParameterTestBlock

from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
)
from watertap.unit_models.nanofiltration_0D import (
    Nanofiltration0D,
    Nanofiltration0DInitializer,
    Nanofiltration0DScaler,
)


class TestBuild:
    @pytest.mark.unit
    def test_config(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -1},
        )
        m.fs.unit = Nanofiltration0D(property_package=m.fs.properties)

        assert len(m.fs.unit.config) == 11
        assert not m.fs.unit.config.dynamic
        assert not m.fs.unit.config.has_holdup
        assert m.fs.unit.config.property_package is m.fs.properties
        assert m.fs.unit.config.property_package_args == {}
        assert (
            m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
        )
        assert not m.fs.unit.config.has_pressure_change
        assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.isothermal
        assert m.fs.unit.config.electroneutrality_ion == "Cl_-"
        assert m.fs.unit.config.passing_species_list is None
        assert m.fs.unit.config.default_passing_rejection == pytest.approx(
            0.1, rel=1e-12
        )
        assert m.fs.unit.config.default_excluded_rejection == pytest.approx(
            1 - 1e-10, rel=1e-12
        )

    @pytest.mark.unit
    def test_default_initializer_and_scaler(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -1},
        )
        m.fs.unit = Nanofiltration0D(property_package=m.fs.properties)

        assert m.fs.unit.default_initializer is Nanofiltration0DInitializer
        assert m.fs.unit.default_scaler is Nanofiltration0DScaler

    @pytest.mark.unit
    def test_multiphase(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(
            ConfigurationError,
            match="Nanofiltration0D model only supports single phase "
            "property packages.",
        ):
            m.fs.unit = Nanofiltration0D(property_package=m.fs.properties)

    @pytest.mark.unit
    def test_electroneutrality_invalid_component(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -1},
        )

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "electroneutrality_ion (foo) "
                "is not a valid component in property package."
            ),
        ):
            m.fs.unit = Nanofiltration0D(
                property_package=m.fs.properties,
                electroneutrality_ion="foo",
            )

    @pytest.mark.unit
    def test_electroneutrality_not_in_passing_list(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -1},
        )

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "electroneutrality_ion (Na_+) "
                "must be a member of the passing species list."
            ),
        ):
            m.fs.unit = Nanofiltration0D(
                property_package=m.fs.properties,
                electroneutrality_ion="Na_+",
                passing_species_list=["Cl_-"],
            )

    @pytest.mark.unit
    def test_electroneutrality_divalent(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -2},
        )

        with pytest.raises(
            ConfigurationError,
            match=re.escape("electroneutrality_ion (Cl_-) must be a monovalent ion."),
        ):
            m.fs.unit = Nanofiltration0D(
                property_package=m.fs.properties,
            )

    unsupported_balance_types = [
        MomentumBalanceType.pressurePhase,
        MomentumBalanceType.momentumTotal,
        MomentumBalanceType.momentumPhase,
    ]

    @pytest.mark.unit
    @pytest.mark.parametrize("balance_type", unsupported_balance_types)
    def test_unsupported_momentum_balance(self, balance_type):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -1},
        )

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                f"Nanofiltration0D model only supports total pressure balance or no pressure balance "
                f"(assigned {balance_type})"
            ),
        ):
            m.fs.unit = Nanofiltration0D(
                property_package=m.fs.properties,
                momentum_balance_type=balance_type,
            )

    @pytest.mark.unit
    def test_invalid_pressure_change(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -1},
        )

        with pytest.raises(
            ConfigurationError,
            match="Inconsistent configuration arguments. has_pressure_change=True "
            "requires that momentum_balance_type not equal MomentumBalanceType.none.",
        ):
            m.fs.unit = Nanofiltration0D(
                property_package=m.fs.properties,
                momentum_balance_type=MomentumBalanceType.none,
                has_pressure_change=True,
            )

    unsupported_ebalance_types = [
        e
        for e in EnergyBalanceType
        if e not in (EnergyBalanceType.none, EnergyBalanceType.isothermal)
    ]

    @pytest.mark.unit
    @pytest.mark.parametrize("balance_type", unsupported_ebalance_types)
    def test_unsupported_energy_balance(self, balance_type):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            charge={"Na_+": 1, "Cl_-": -1},
        )

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                f"Nanofiltration0D model only supports isothermal operation or no energy balance "
                f"(assigned {balance_type})"
            ),
        ):
            m.fs.unit = Nanofiltration0D(
                property_package=m.fs.properties,
                energy_balance_type=balance_type,
            )

    @pytest.mark.unit
    def test_default_build(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
        )
        m.fs.unit = Nanofiltration0D(property_package=m.fs.properties)

        assert hasattr(m.fs.unit, "properties_in")
        assert hasattr(m.fs.unit, "properties_retentate")
        assert hasattr(m.fs.unit, "properties_permeate")

        assert isinstance(m.fs.unit.inlet, Port)
        assert isinstance(m.fs.unit.retentate, Port)
        assert isinstance(m.fs.unit.permeate, Port)

        assert isinstance(m.fs.unit._solute_set, Set)
        assert len(m.fs.unit._solute_set) == 4
        for j in m.fs.unit._solute_set:
            assert j in ["Na_+", "Ca_2+", "SO4_2-", "Mg_2+"]

        assert isinstance(m.fs.unit.recovery_solvent, Var)
        assert value(m.fs.unit.recovery_solvent[0]) == pytest.approx(0.8, rel=1e-8)

        assert isinstance(m.fs.unit.rejection_comp, Var)
        for (_, j), v in m.fs.unit.rejection_comp.items():
            if j in ["Ca_2+", "SO4_2-", "Mg_2+"]:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_excluded_rejection, abs=1e-12
                )
            else:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_passing_rejection, abs=1e-12
                )

        assert not hasattr(m.fs.unit, "deltaP")

        assert isinstance(m.fs.unit.material_balances, Constraint)
        assert len(m.fs.unit.material_balances) == 6  # 1 time point, 6 components
        for (t, j), cd in m.fs.unit.material_balances.items():
            assert str(cd.expr) == str(
                m.fs.unit.properties_in[t].flow_mol_phase_comp["Liq", j]
                == m.fs.unit.properties_retentate[t].flow_mol_phase_comp["Liq", j]
                + m.fs.unit.properties_permeate[t].flow_mol_phase_comp["Liq", j]
            )

        assert isinstance(m.fs.unit.separation_constraint, Constraint)
        assert (
            len(m.fs.unit.separation_constraint) == 5
        )  # 1 time point, 5 components (Cl- is skipped)
        for (t, p, j), cd in m.fs.unit.separation_constraint.items():
            assert j != "Cl_-"

            if j == "H2O":
                assert str(cd.expr) == str(
                    m.fs.unit.recovery_solvent[t]
                    * m.fs.unit.properties_in[t].flow_mol_phase_comp[p, j]
                    == m.fs.unit.properties_permeate[t].flow_mol_phase_comp[p, j]
                )
            else:
                assert str(cd.expr) == str(
                    (
                        (1 - m.fs.unit.rejection_comp[t, j])
                        * m.fs.unit.properties_in[t].conc_mol_phase_comp[p, j]
                    )
                    == m.fs.unit.properties_permeate[t].conc_mol_phase_comp[p, j]
                )

        assert isinstance(m.fs.unit.permeate_electronegativity, Constraint)
        assert len(m.fs.unit.permeate_electronegativity) == 1
        assert str(m.fs.unit.permeate_electronegativity[0].expr) == str(
            0
            == sum(
                m.fs.properties.get_component(j).config.charge
                * m.fs.unit.properties_permeate[0].flow_mol_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        )

        assert isinstance(m.fs.unit.retentate_pressure_balance, Constraint)
        assert len(m.fs.unit.retentate_pressure_balance) == 1
        assert str(m.fs.unit.retentate_pressure_balance[0].expr) == str(
            m.fs.unit.properties_in[0].pressure
            == m.fs.unit.properties_retentate[0].pressure
        )

        assert isinstance(m.fs.unit.retentate_temperature_equality, Constraint)
        assert len(m.fs.unit.retentate_temperature_equality) == 1
        assert str(m.fs.unit.retentate_temperature_equality[0].expr) == str(
            m.fs.unit.properties_in[0].temperature
            == m.fs.unit.properties_retentate[0].temperature
        )

        assert isinstance(m.fs.unit.permeate_temperature_equality, Constraint)
        assert len(m.fs.unit.permeate_temperature_equality) == 1
        assert str(m.fs.unit.permeate_temperature_equality[0].expr) == str(
            m.fs.unit.properties_in[0].temperature
            == m.fs.unit.properties_permeate[0].temperature
        )

    @pytest.mark.unit
    def test_w_passing_listd(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
        )
        m.fs.unit = Nanofiltration0D(
            property_package=m.fs.properties,
            passing_species_list=["Mg_2+", "Na_+", "Cl_-"],
        )

        assert hasattr(m.fs.unit, "properties_in")
        assert hasattr(m.fs.unit, "properties_retentate")
        assert hasattr(m.fs.unit, "properties_permeate")

        assert isinstance(m.fs.unit.inlet, Port)
        assert isinstance(m.fs.unit.retentate, Port)
        assert isinstance(m.fs.unit.permeate, Port)

        assert isinstance(m.fs.unit._solute_set, Set)
        assert len(m.fs.unit._solute_set) == 4
        for j in m.fs.unit._solute_set:
            assert j in ["Na_+", "Ca_2+", "SO4_2-", "Mg_2+"]

        assert isinstance(m.fs.unit.recovery_solvent, Var)
        assert value(m.fs.unit.recovery_solvent[0]) == pytest.approx(0.8, rel=1e-8)

        assert isinstance(m.fs.unit.rejection_comp, Var)
        for (_, j), v in m.fs.unit.rejection_comp.items():
            if j in ["Ca_2+", "SO4_2-"]:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_excluded_rejection, abs=1e-12
                )
            else:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_passing_rejection, abs=1e-12
                )

        assert not hasattr(m.fs.unit, "deltaP")

        assert isinstance(m.fs.unit.material_balances, Constraint)
        assert len(m.fs.unit.material_balances) == 6  # 1 time point, 6 components
        for (t, j), cd in m.fs.unit.material_balances.items():
            assert str(cd.expr) == str(
                m.fs.unit.properties_in[t].flow_mol_phase_comp["Liq", j]
                == m.fs.unit.properties_retentate[t].flow_mol_phase_comp["Liq", j]
                + m.fs.unit.properties_permeate[t].flow_mol_phase_comp["Liq", j]
            )

        assert isinstance(m.fs.unit.separation_constraint, Constraint)
        assert (
            len(m.fs.unit.separation_constraint) == 5
        )  # 1 time point, 5 components (Cl- is skipped)
        for (t, p, j), cd in m.fs.unit.separation_constraint.items():
            assert j != "Cl_-"

            if j == "H2O":
                assert str(cd.expr) == str(
                    m.fs.unit.recovery_solvent[t]
                    * m.fs.unit.properties_in[t].flow_mol_phase_comp[p, j]
                    == m.fs.unit.properties_permeate[t].flow_mol_phase_comp[p, j]
                )
            else:
                assert str(cd.expr) == str(
                    (
                        (1 - m.fs.unit.rejection_comp[t, j])
                        * m.fs.unit.properties_in[t].conc_mol_phase_comp[p, j]
                    )
                    == m.fs.unit.properties_permeate[t].conc_mol_phase_comp[p, j]
                )

        assert isinstance(m.fs.unit.permeate_electronegativity, Constraint)
        assert len(m.fs.unit.permeate_electronegativity) == 1
        assert str(m.fs.unit.permeate_electronegativity[0].expr) == str(
            0
            == sum(
                m.fs.properties.get_component(j).config.charge
                * m.fs.unit.properties_permeate[0].flow_mol_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        )

        assert isinstance(m.fs.unit.retentate_pressure_balance, Constraint)
        assert len(m.fs.unit.retentate_pressure_balance) == 1
        assert str(m.fs.unit.retentate_pressure_balance[0].expr) == str(
            m.fs.unit.properties_in[0].pressure
            == m.fs.unit.properties_retentate[0].pressure
        )

        assert isinstance(m.fs.unit.retentate_temperature_equality, Constraint)
        assert len(m.fs.unit.retentate_temperature_equality) == 1
        assert str(m.fs.unit.retentate_temperature_equality[0].expr) == str(
            m.fs.unit.properties_in[0].temperature
            == m.fs.unit.properties_retentate[0].temperature
        )

        assert isinstance(m.fs.unit.permeate_temperature_equality, Constraint)
        assert len(m.fs.unit.permeate_temperature_equality) == 1
        assert str(m.fs.unit.permeate_temperature_equality[0].expr) == str(
            m.fs.unit.properties_in[0].temperature
            == m.fs.unit.properties_permeate[0].temperature
        )

    @pytest.mark.unit
    def test_build_w_pressure_drop(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
        )
        m.fs.unit = Nanofiltration0D(
            property_package=m.fs.properties,
            has_pressure_change=True,
        )

        assert hasattr(m.fs.unit, "properties_in")
        assert hasattr(m.fs.unit, "properties_retentate")
        assert hasattr(m.fs.unit, "properties_permeate")

        assert isinstance(m.fs.unit.inlet, Port)
        assert isinstance(m.fs.unit.retentate, Port)
        assert isinstance(m.fs.unit.permeate, Port)

        assert isinstance(m.fs.unit._solute_set, Set)
        assert len(m.fs.unit._solute_set) == 4
        for j in m.fs.unit._solute_set:
            assert j in ["Na_+", "Ca_2+", "SO4_2-", "Mg_2+"]

        assert isinstance(m.fs.unit.recovery_solvent, Var)
        assert value(m.fs.unit.recovery_solvent[0]) == pytest.approx(0.8, rel=1e-8)

        assert isinstance(m.fs.unit.rejection_comp, Var)
        for (_, j), v in m.fs.unit.rejection_comp.items():
            if j in ["Ca_2+", "SO4_2-", "Mg_2+"]:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_excluded_rejection, abs=1e-12
                )
            else:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_passing_rejection, abs=1e-12
                )

        assert isinstance(m.fs.unit.deltaP, Var)

        assert isinstance(m.fs.unit.material_balances, Constraint)
        assert len(m.fs.unit.material_balances) == 6  # 1 time point, 6 components
        for (t, j), cd in m.fs.unit.material_balances.items():
            assert str(cd.expr) == str(
                m.fs.unit.properties_in[t].flow_mol_phase_comp["Liq", j]
                == m.fs.unit.properties_retentate[t].flow_mol_phase_comp["Liq", j]
                + m.fs.unit.properties_permeate[t].flow_mol_phase_comp["Liq", j]
            )

        assert isinstance(m.fs.unit.separation_constraint, Constraint)
        assert (
            len(m.fs.unit.separation_constraint) == 5
        )  # 1 time point, 5 components (Cl- is skipped)
        for (t, p, j), cd in m.fs.unit.separation_constraint.items():
            assert j != "Cl_-"

            if j == "H2O":
                assert str(cd.expr) == str(
                    m.fs.unit.recovery_solvent[t]
                    * m.fs.unit.properties_in[t].flow_mol_phase_comp[p, j]
                    == m.fs.unit.properties_permeate[t].flow_mol_phase_comp[p, j]
                )
            else:
                assert str(cd.expr) == str(
                    (
                        (1 - m.fs.unit.rejection_comp[t, j])
                        * m.fs.unit.properties_in[t].conc_mol_phase_comp[p, j]
                    )
                    == m.fs.unit.properties_permeate[t].conc_mol_phase_comp[p, j]
                )

        assert isinstance(m.fs.unit.permeate_electronegativity, Constraint)
        assert len(m.fs.unit.permeate_electronegativity) == 1
        assert str(m.fs.unit.permeate_electronegativity[0].expr) == str(
            0
            == sum(
                m.fs.properties.get_component(j).config.charge
                * m.fs.unit.properties_permeate[0].flow_mol_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        )

        assert isinstance(m.fs.unit.retentate_pressure_balance, Constraint)
        assert len(m.fs.unit.retentate_pressure_balance) == 1
        assert str(m.fs.unit.retentate_pressure_balance[0].expr) == str(
            m.fs.unit.properties_in[0].pressure
            == m.fs.unit.properties_retentate[0].pressure - m.fs.unit.deltaP[0]
        )

        assert isinstance(m.fs.unit.retentate_temperature_equality, Constraint)
        assert len(m.fs.unit.retentate_temperature_equality) == 1
        assert str(m.fs.unit.retentate_temperature_equality[0].expr) == str(
            m.fs.unit.properties_in[0].temperature
            == m.fs.unit.properties_retentate[0].temperature
        )

        assert isinstance(m.fs.unit.permeate_temperature_equality, Constraint)
        assert len(m.fs.unit.permeate_temperature_equality) == 1
        assert str(m.fs.unit.permeate_temperature_equality[0].expr) == str(
            m.fs.unit.properties_in[0].temperature
            == m.fs.unit.properties_permeate[0].temperature
        )

    @pytest.mark.unit
    def test_build_no_pt_electroneutrality(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
        )
        m.fs.unit = Nanofiltration0D(
            property_package=m.fs.properties,
            electroneutrality_ion=None,
            momentum_balance_type=MomentumBalanceType.none,
            energy_balance_type=EnergyBalanceType.none,
        )

        assert hasattr(m.fs.unit, "properties_in")
        assert hasattr(m.fs.unit, "properties_retentate")
        assert hasattr(m.fs.unit, "properties_permeate")

        assert isinstance(m.fs.unit.inlet, Port)
        assert isinstance(m.fs.unit.retentate, Port)
        assert isinstance(m.fs.unit.permeate, Port)

        assert isinstance(m.fs.unit._solute_set, Set)
        assert len(m.fs.unit._solute_set) == 5
        for j in m.fs.unit._solute_set:
            assert j in ["Na_+", "Ca_2+", "SO4_2-", "Mg_2+", "Cl_-"]

        assert isinstance(m.fs.unit.recovery_solvent, Var)
        assert value(m.fs.unit.recovery_solvent[0]) == pytest.approx(0.8, rel=1e-8)

        assert isinstance(m.fs.unit.rejection_comp, Var)
        for (_, j), v in m.fs.unit.rejection_comp.items():
            if j in ["Ca_2+", "SO4_2-", "Mg_2+"]:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_excluded_rejection, abs=1e-12
                )
            else:
                assert value(v) == pytest.approx(
                    m.fs.unit.config.default_passing_rejection, abs=1e-12
                )

        assert not hasattr(m.fs.unit, "deltaP")

        assert isinstance(m.fs.unit.material_balances, Constraint)
        assert len(m.fs.unit.material_balances) == 6  # 1 time point, 6 components
        for (t, j), cd in m.fs.unit.material_balances.items():
            assert str(cd.expr) == str(
                m.fs.unit.properties_in[t].flow_mol_phase_comp["Liq", j]
                == m.fs.unit.properties_retentate[t].flow_mol_phase_comp["Liq", j]
                + m.fs.unit.properties_permeate[t].flow_mol_phase_comp["Liq", j]
            )

        assert isinstance(m.fs.unit.separation_constraint, Constraint)
        assert len(m.fs.unit.separation_constraint) == 6  # 1 time point, 6 components
        for (t, p, j), cd in m.fs.unit.separation_constraint.items():
            if j == "H2O":
                assert str(cd.expr) == str(
                    m.fs.unit.recovery_solvent[t]
                    * m.fs.unit.properties_in[t].flow_mol_phase_comp[p, j]
                    == m.fs.unit.properties_permeate[t].flow_mol_phase_comp[p, j]
                )
            else:
                assert str(cd.expr) == str(
                    (
                        (1 - m.fs.unit.rejection_comp[t, j])
                        * m.fs.unit.properties_in[t].conc_mol_phase_comp[p, j]
                    )
                    == m.fs.unit.properties_permeate[t].conc_mol_phase_comp[p, j]
                )

        assert not hasattr(m.fs.unit, "permeate_electronegativity")
        assert not hasattr(m.fs.unit, "retentate_pressure_balance")
        assert not hasattr(m.fs.unit, "retentate_temperature_equality")
        assert not hasattr(m.fs.unit, "permeate_temperature_equality")


@pytest.fixture
def mcas_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(
        solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
        diffusivity_data={
            ("Liq", "Ca_2+"): 9.2e-10,
            ("Liq", "SO4_2-"): 1.06e-09,
            ("Liq", "Mg_2+"): 7.06e-10,
            ("Liq", "Na_+"): 1.33e-09,
            ("Liq", "Cl_-"): 2.03e-09,
        },
        mw_data={
            "H2O": 0.018,
            "Ca_2+": 0.04,
            "Mg_2+": 0.024,
            "SO4_2-": 0.096,
            "Na_+": 0.023,
            "Cl_-": 0.035,
        },
        stokes_radius_data={
            "Ca_2+": 3.09e-10,
            "Mg_2+": 3.47e-10,
            "SO4_2-": 2.3e-10,
            "Cl_-": 1.21e-10,
            "Na_+": 1.84e-10,
        },
        charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
    )

    m.fs.unit = Nanofiltration0D(
        property_package=m.fs.properties,
        has_pressure_change=True,
    )

    return m


class TestNanofiltration0DInitializer:

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_initialize(self, mcas_model):
        # Fix DoF
        mcas_model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(53.6036)
        mcas_model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix(0.00955)
        mcas_model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Mg_2+"].fix(0.5808)
        mcas_model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "SO4_2-"].fix(0.02225)
        mcas_model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(1.61977)
        mcas_model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.48357)
        mcas_model.fs.unit.inlet.temperature[0].fix(298.15)
        mcas_model.fs.unit.inlet.pressure[0].fix(4e5)

        mcas_model.fs.unit.recovery_solvent.fix()
        mcas_model.fs.unit.rejection_comp.fix()
        mcas_model.fs.unit.permeate.pressure[0].fix(1e5)
        mcas_model.fs.unit.deltaP.fix(-1e4)
        mcas_model.fs.unit.area.fix(500)

        # Initialize
        initializer = Nanofiltration0DInitializer()

        initializer.initialize(mcas_model.fs.unit)

        assert (
            initializer.summary[mcas_model.fs.unit]["status"] == InitializationStatus.Ok
        )

        for c in mcas_model.component_data_objects(Constraint, descend_into=True):
            assert (
                abs(value(c.body - c.lower)) <= 1e-5
                and abs(value(c.upper - c.body)) <= 1e-5
            )

    @pytest.mark.unit
    def test_init_conc(self, mcas_model):
        # Use MCAS model to make dummy calls to initializer sub methods
        initializer = Nanofiltration0DInitializer()

        mcas_model.in_var = Var(mcas_model.fs.properties.component_list, initialize=10)
        mcas_model.ret_var = Var(mcas_model.fs.properties.component_list, initialize=0)
        mcas_model.perm_var = Var(mcas_model.fs.properties.component_list, initialize=0)

        initializer._init_conc(
            mcas_model.fs.unit,
            0,  # time index
            mcas_model.fs.unit.properties_in[0],
            mcas_model.in_var,
            mcas_model.ret_var,
            mcas_model.perm_var,
        )

        assert value(mcas_model.ret_var["H2O"]) == pytest.approx(
            10, rel=1e-10
        )  # solvent concentration unchanged
        assert value(mcas_model.ret_var["Cl_-"]) == pytest.approx(
            2, rel=1e-10
        )  # should use solvent recovery
        assert value(mcas_model.ret_var["Na_+"]) == pytest.approx(1, rel=1e-10)
        assert value(mcas_model.ret_var["Ca_2+"]) == pytest.approx(
            10 * (1 - 1e-10), rel=1e-10
        )
        assert value(mcas_model.ret_var["Mg_2+"]) == pytest.approx(
            10 * (1 - 1e-10), rel=1e-10
        )
        assert value(mcas_model.ret_var["SO4_2-"]) == pytest.approx(
            10 * (1 - 1e-10), rel=1e-10
        )

        assert value(mcas_model.perm_var["H2O"]) == pytest.approx(
            10, rel=1e-10
        )  # solvent concentration unchanged
        assert value(mcas_model.perm_var["Cl_-"]) == pytest.approx(
            8, rel=1e-10
        )  # should use solvent recovery
        assert value(mcas_model.perm_var["Na_+"]) == pytest.approx(9, rel=1e-10)
        assert value(mcas_model.perm_var["Ca_2+"]) == pytest.approx(
            10 * 1e-10, rel=1e-10
        )
        assert value(mcas_model.perm_var["Mg_2+"]) == pytest.approx(
            10 * 1e-10, rel=1e-10
        )
        assert value(mcas_model.perm_var["SO4_2-"]) == pytest.approx(
            10 * 1e-10, rel=1e-10
        )

    @pytest.mark.unit
    def test_init_flow_comp(self, mcas_model):
        # Use MCAS model to make dummy calls to initializer sub methods
        initializer = Nanofiltration0DInitializer()

        mcas_model.in_var_comp = Var(
            mcas_model.fs.properties.component_list, initialize=10
        )
        mcas_model.ret_var = Var(mcas_model.fs.properties.component_list, initialize=0)
        mcas_model.perm_var = Var(mcas_model.fs.properties.component_list, initialize=0)

        initializer._init_flow(
            mcas_model.fs.unit,
            0,  # time index
            mcas_model.fs.unit.properties_in[0],
            mcas_model.in_var_comp,
            mcas_model.ret_var,
            mcas_model.perm_var,
        )

        assert value(mcas_model.ret_var["H2O"]) == pytest.approx(2, rel=1e-10)
        assert value(mcas_model.ret_var["Cl_-"]) == pytest.approx(
            2, rel=1e-10
        )  # should use solvent recovery
        assert value(mcas_model.ret_var["Na_+"]) == pytest.approx(1, rel=1e-10)
        assert value(mcas_model.ret_var["Ca_2+"]) == pytest.approx(
            10 * (1 - 1e-10), rel=1e-10
        )
        assert value(mcas_model.ret_var["Mg_2+"]) == pytest.approx(
            10 * (1 - 1e-10), rel=1e-10
        )
        assert value(mcas_model.ret_var["SO4_2-"]) == pytest.approx(
            10 * (1 - 1e-10), rel=1e-10
        )

        assert value(mcas_model.perm_var["H2O"]) == pytest.approx(8, rel=1e-10)
        assert value(mcas_model.perm_var["Cl_-"]) == pytest.approx(
            8, rel=1e-10
        )  # should use solvent recovery
        assert value(mcas_model.perm_var["Na_+"]) == pytest.approx(9, rel=1e-10)
        assert value(mcas_model.perm_var["Ca_2+"]) == pytest.approx(
            10 * 1e-10, rel=1e-10
        )
        assert value(mcas_model.perm_var["Mg_2+"]) == pytest.approx(
            10 * 1e-10, rel=1e-10
        )
        assert value(mcas_model.perm_var["SO4_2-"]) == pytest.approx(
            10 * 1e-10, rel=1e-10
        )

    @pytest.mark.unit
    def test_init_flow_vol(self, mcas_model):
        # Use MCAS model to make dummy calls to initializer sub methods
        initializer = Nanofiltration0DInitializer()

        mcas_model.in_var_vol = Var(initialize=10)
        mcas_model.ret_var = Var(initialize=0)
        mcas_model.perm_var = Var(initialize=0)

        initializer._init_flow(
            mcas_model.fs.unit,
            0,  # time index
            mcas_model.fs.unit.properties_in[0],
            mcas_model.in_var_vol,
            mcas_model.ret_var,
            mcas_model.perm_var,
        )

        assert value(mcas_model.ret_var) == pytest.approx(2, rel=1e-10)
        assert value(mcas_model.perm_var) == pytest.approx(8, rel=1e-10)

    @pytest.mark.unit
    def test_init_frac(self, mcas_model):
        # Use MCAS model to make dummy calls to initializer sub methods
        initializer = Nanofiltration0DInitializer()

        mcas_model.in_var = Var(mcas_model.fs.properties.component_list, initialize=10)
        mcas_model.ret_var = Var(mcas_model.fs.properties.component_list, initialize=0)
        mcas_model.perm_var = Var(mcas_model.fs.properties.component_list, initialize=0)

        initializer._init_frac(
            mcas_model.fs.unit,
            0,  # time index
            mcas_model.fs.unit.properties_in[0],
            mcas_model.in_var,
            mcas_model.ret_var,
            mcas_model.perm_var,
        )

        assert value(mcas_model.ret_var["H2O"]) == pytest.approx(2 / 35, rel=1e-6)
        assert value(mcas_model.ret_var["Cl_-"]) == pytest.approx(2 / 35, rel=1e-6)
        assert value(mcas_model.ret_var["Na_+"]) == pytest.approx(1 / 35, rel=1e-6)
        assert value(mcas_model.ret_var["Ca_2+"]) == pytest.approx(10 / 35, rel=1e-6)
        assert value(mcas_model.ret_var["Mg_2+"]) == pytest.approx(10 / 35, rel=1e-6)
        assert value(mcas_model.ret_var["SO4_2-"]) == pytest.approx(10 / 35, rel=1e-6)

        assert value(mcas_model.perm_var["H2O"]) == pytest.approx(8 / 25, rel=1e-6)
        assert value(mcas_model.perm_var["Cl_-"]) == pytest.approx(8 / 25, rel=1e-6)
        assert value(mcas_model.perm_var["Na_+"]) == pytest.approx(9 / 25, rel=1e-6)
        assert value(mcas_model.perm_var["Ca_2+"]) == pytest.approx(0, abs=1e-6)
        assert value(mcas_model.perm_var["Mg_2+"]) == pytest.approx(0, abs=1e-6)
        assert value(mcas_model.perm_var["SO4_2-"]) == pytest.approx(0, abs=1e-6)


class TestNanofiltration0DScaler:
    @pytest.mark.unit
    def test_variable_scaling_routine(self, mcas_model):
        scaler = Nanofiltration0DScaler()

        scaler.variable_scaling_routine(mcas_model.fs.unit)

        # Check that submodels have scaling factors
        assert isinstance(mcas_model.fs.unit.properties_in[0].scaling_factor, Suffix)
        assert len(mcas_model.fs.unit.properties_in[0].scaling_factor) > 1

        assert isinstance(
            mcas_model.fs.unit.properties_retentate[0].scaling_factor, Suffix
        )
        assert len(mcas_model.fs.unit.properties_retentate[0].scaling_factor) > 1

        assert isinstance(
            mcas_model.fs.unit.properties_permeate[0].scaling_factor, Suffix
        )
        assert len(mcas_model.fs.unit.properties_permeate[0].scaling_factor) > 1

        # Check scaling factors for unit variables
        assert isinstance(mcas_model.fs.unit.scaling_factor, Suffix)
        assert len(mcas_model.fs.unit.scaling_factor) == 6

        assert mcas_model.fs.unit.scaling_factor[
            mcas_model.fs.unit.deltaP[0]
        ] == pytest.approx(1e-4, rel=1e-8)

        for j in ["Ca_2+", "Mg_2+", "SO4_2-", "Na_+"]:
            assert (
                mcas_model.fs.unit.scaling_factor[
                    mcas_model.fs.unit.rejection_comp[0, j]
                ]
                == 1e2
            )
        assert (
            mcas_model.fs.unit.scaling_factor[mcas_model.fs.unit.recovery_solvent[0]]
            == 10
        )

    @pytest.mark.unit
    def test_constraint_scaling_routine(self, mcas_model):
        scaler = Nanofiltration0DScaler()

        scaler.constraint_scaling_routine(mcas_model.fs.unit)

        # Check that submodels have scaling factors
        # No constraint scaling factors applied ot sub-models, so no check for length of Suffix
        assert isinstance(mcas_model.fs.unit.properties_in[0].scaling_factor, Suffix)
        assert isinstance(
            mcas_model.fs.unit.properties_retentate[0].scaling_factor, Suffix
        )
        assert isinstance(
            mcas_model.fs.unit.properties_permeate[0].scaling_factor, Suffix
        )

        # Check scaling factors for unit variables
        assert isinstance(mcas_model.fs.unit.scaling_factor, Suffix)
        assert len(mcas_model.fs.unit.scaling_factor) == 15

        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.material_balances[0.0, "Ca_2+"]
            ]
            == 10.0
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.material_balances[0.0, "Cl_-"]
            ]
            == 10.0
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.material_balances[0.0, "H2O"]
            ]
            == 10.0
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.material_balances[0.0, "Mg_2+"]
            ]
            == 10.0
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.material_balances[0.0, "Na_+"]
            ]
            == 10.0
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.material_balances[0.0, "SO4_2-"]
            ]
            == 10.0
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.permeate_electronegativity[0.0]
            ]
            == 5.0
        )
        assert mcas_model.fs.unit.scaling_factor[
            mcas_model.fs.unit.permeate_temperature_equality[0.0]
        ] == pytest.approx(3.354016e-3, rel=1e-5)

        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.separation_constraint[0.0, "Liq", "Ca_2+"]
            ]
            == 0.002
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.separation_constraint[0.0, "Liq", "H2O"]
            ]
            == 10.0
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.separation_constraint[0.0, "Liq", "Mg_2+"]
            ]
            == 0.002
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.separation_constraint[0.0, "Liq", "Na_+"]
            ]
            == 0.002
        )
        assert (
            mcas_model.fs.unit.scaling_factor[
                mcas_model.fs.unit.separation_constraint[0.0, "Liq", "SO4_2-"]
            ]
            == 0.002
        )
        assert mcas_model.fs.unit.scaling_factor[
            mcas_model.fs.unit.retentate_pressure_balance[0.0]
        ] == pytest.approx(9.869233e-06, rel=1e-5)
        assert mcas_model.fs.unit.scaling_factor[
            mcas_model.fs.unit.retentate_temperature_equality[0.0]
        ] == pytest.approx(3.354016e-3, rel=1e-5)


class TestMCAS:
    @pytest.fixture(scope="class")
    def mcas_case(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            diffusivity_data={
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "SO4_2-"): 1.06e-09,
                ("Liq", "Mg_2+"): 7.06e-10,
                ("Liq", "Na_+"): 1.33e-09,
                ("Liq", "Cl_-"): 2.03e-09,
            },
            mw_data={
                "H2O": 0.018,
                "Ca_2+": 0.04,
                "Mg_2+": 0.024,
                "SO4_2-": 0.096,
                "Na_+": 0.023,
                "Cl_-": 0.035,
            },
            stokes_radius_data={
                "Ca_2+": 3.09e-10,
                "Mg_2+": 3.47e-10,
                "SO4_2-": 2.3e-10,
                "Cl_-": 1.21e-10,
                "Na_+": 1.84e-10,
            },
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
        )

        m.fs.unit = Nanofiltration0D(
            property_package=m.fs.properties,
            has_pressure_change=True,
        )

        # Fix other inlet state variables
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(53.6036)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix(0.00955)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Mg_2+"].fix(0.5808)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "SO4_2-"].fix(0.02225)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(1.61977)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.48357)
        m.fs.unit.inlet.temperature[0].fix(298.15)
        m.fs.unit.inlet.pressure[0].fix(4e5)

        m.fs.unit.recovery_solvent.fix()
        m.fs.unit.rejection_comp.fix()
        m.fs.unit.area.fix(500)
        m.fs.unit.deltaP.fix(-1e4)
        m.fs.unit.permeate.pressure[0].fix(101325)

        return m

    @pytest.mark.component
    def test_structural_issues(self, mcas_case):
        dt = DiagnosticsToolbox(mcas_case)

        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.component
    def test_scaling(self, mcas_case):
        # Scale model
        set_scaling_factor(
            mcas_case.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"], 1e-1
        )
        set_scaling_factor(
            mcas_case.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"], 1e3
        )
        set_scaling_factor(
            mcas_case.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Mg_2+"], 10
        )
        set_scaling_factor(
            mcas_case.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "SO4_2-"], 1e2
        )
        set_scaling_factor(
            mcas_case.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"], 1
        )
        set_scaling_factor(
            mcas_case.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"], 10
        )

        scaler = Nanofiltration0DScaler()
        scaler.scale_model(mcas_case.fs.unit)

        assert (
            mcas_case.fs.unit.scaling_factor[mcas_case.fs.unit.recovery_solvent[0]]
            == 10
        )
        for j in ["Ca_2+", "Mg_2+", "SO4_2-", "Na_+"]:
            assert (
                mcas_case.fs.unit.scaling_factor[mcas_case.fs.unit.rejection_comp[0, j]]
                == 100
            )

        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.deltaP[0.0]
        ] == pytest.approx(0.0001, rel=1e-5)

        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.material_balances[0.0, "H2O"]
        ] == pytest.approx(0.1, rel=1e-3)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.material_balances[0.0, "Ca_2+"]
        ] == pytest.approx(1e3, rel=1e-3)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.material_balances[0.0, "SO4_2-"]
        ] == pytest.approx(1e2, rel=1e-3)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.material_balances[0.0, "Mg_2+"]
        ] == pytest.approx(10, rel=1e-3)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.material_balances[0.0, "Na_+"]
        ] == pytest.approx(10, rel=1e-3)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.material_balances[0.0, "Cl_-"]
        ] == pytest.approx(1, rel=1e-3)

        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.separation_constraint[0.0, "Liq", "H2O"]
        ] == pytest.approx(0.1, rel=1e-5)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.separation_constraint[0.0, "Liq", "Ca_2+"]
        ] == pytest.approx(0.004, rel=1e-5)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.separation_constraint[0.0, "Liq", "SO4_2-"]
        ] == pytest.approx(0.00960, rel=1e-5)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.separation_constraint[0.0, "Liq", "Mg_2+"]
        ] == pytest.approx(0.00240, rel=1e-5)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.separation_constraint[0.0, "Liq", "Na_+"]
        ] == pytest.approx(0.0023, rel=1e-3)

        assert (
            mcas_case.fs.unit.scaling_factor[
                mcas_case.fs.unit.permeate_electronegativity[0.0]
            ]
            == 1
        )
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.retentate_pressure_balance[0.0]
        ] == pytest.approx(1e-5, rel=1e-3)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.retentate_temperature_equality[0.0]
        ] == pytest.approx(1e-2, rel=1e-3)
        assert mcas_case.fs.unit.scaling_factor[
            mcas_case.fs.unit.permeate_temperature_equality[0.0]
        ] == pytest.approx(1e-2, rel=1e-3)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_initialize(self, mcas_case):
        initializer = Nanofiltration0DInitializer()
        initializer.initialize(mcas_case.fs.unit)

        assert (
            initializer.summary[mcas_case.fs.unit]["status"] == InitializationStatus.Ok
        )

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solve(self, mcas_case):
        solver = get_solver(
            "ipopt_v2", writer_config={"scale_model": True, "linear_presolve": True}
        )
        results = solver.solve(mcas_case)

        assert_optimal_termination(results)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_numerical_issues(self, mcas_case):
        sm = TransformationFactory("core.scale_model").create_using(
            mcas_case, rename=False
        )

        dt = DiagnosticsToolbox(sm.fs.unit)
        dt.assert_no_numerical_warnings()

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solution(self, mcas_case):
        # Retentate stream
        assert value(
            mcas_case.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "H2O"]
        ) == pytest.approx(0.2 * 53.6036, rel=1e-5)
        assert value(
            mcas_case.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Ca_2+"]
        ) == pytest.approx(0.00955, rel=1e-5)
        assert value(
            mcas_case.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Mg_2+"]
        ) == pytest.approx(0.5808, rel=1e-5)
        assert value(
            mcas_case.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "SO4_2-"]
        ) == pytest.approx(0.02225, rel=1e-5)
        assert value(
            mcas_case.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Cl_-"]
        ) == pytest.approx(1.29167, rel=1e-5)
        assert value(
            mcas_case.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Na_+"]
        ) == pytest.approx(0.155472, rel=1e-5)
        assert value(mcas_case.fs.unit.retentate.temperature[0]) == pytest.approx(
            298.15, rel=1e-5
        )
        assert value(mcas_case.fs.unit.retentate.pressure[0]) == pytest.approx(
            3.9e5, rel=1e-5
        )

        # Permeate stream
        assert value(
            mcas_case.fs.unit.permeate.flow_mol_phase_comp[0, "Liq", "H2O"]
        ) == pytest.approx(0.8 * 53.6036, rel=1e-5)
        assert value(
            mcas_case.fs.unit.permeate.flow_mol_phase_comp[0, "Liq", "Ca_2+"]
        ) == pytest.approx(0, abs=1e-7)
        assert value(
            mcas_case.fs.unit.permeate.flow_mol_phase_comp[0, "Liq", "Mg_2+"]
        ) == pytest.approx(0, abs=1e-7)
        assert value(
            mcas_case.fs.unit.permeate.flow_mol_phase_comp[0, "Liq", "SO4_2-"]
        ) == pytest.approx(0, abs=1e-7)
        assert value(
            mcas_case.fs.unit.permeate.flow_mol_phase_comp[0, "Liq", "Cl_-"]
        ) == pytest.approx(0.328098, rel=1e-5)
        assert value(
            mcas_case.fs.unit.permeate.flow_mol_phase_comp[0, "Liq", "Na_+"]
        ) == pytest.approx(0.328098, rel=1e-5)
        assert value(mcas_case.fs.unit.permeate.temperature[0]) == pytest.approx(
            298.15, rel=1e-5
        )
        assert value(mcas_case.fs.unit.permeate.pressure[0]) == pytest.approx(
            101325, rel=1e-5
        )
        assert value(mcas_case.fs.unit.recovery_vol_phase[0, "Liq"]) == pytest.approx(
            0.753868, rel=1e-3
        )
        assert value(
            mcas_case.fs.unit.recovery_mass_phase_comp[0, "Liq", "H2O"]
        ) == pytest.approx(0.79999, rel=1e-3)
        assert value(
            mcas_case.fs.unit.recovery_mass_phase_comp[0, "Liq", "Na_+"]
        ) == pytest.approx(0.67848, rel=1e-3)

        # Rejection
        for j in ["Ca_2+", "Mg_2+", "SO4_2-"]:
            assert value(
                mcas_case.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", j]
                / mcas_case.fs.unit.properties_in[0].conc_mol_phase_comp["Liq", j]
            ) == pytest.approx(0, abs=1e-8)
        assert value(
            mcas_case.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Na_+"]
            / mcas_case.fs.unit.properties_in[0].conc_mol_phase_comp["Liq", "Na_+"]
        ) == pytest.approx(0.9, abs=1e-8)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_conservation(self, mcas_case):
        for j in mcas_case.fs.properties.component_list:
            assert (
                abs(
                    value(
                        mcas_case.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", j]
                        - mcas_case.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", j]
                        - mcas_case.fs.unit.permeate.flow_mol_phase_comp[0, "Liq", j]
                    )
                )
                <= 1e-6
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order granular activated carbon model
"""
import pytest


from pyomo.environ import (
    Block,
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import GACZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestGACZO_w_o_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tss", "nonvolatile_toc"])

        m.fs.unit = GACZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.empty_bed_contact_time, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.activated_carbon_replacement, Var)
        assert isinstance(model.fs.unit.activated_carbon_demand, Var)
        assert isinstance(model.fs.unit.activated_carbon_bulk_density, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.electricity_intensity_constraint, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)
        assert isinstance(model.fs.unit.activated_carbon_equation, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("gac")

        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.empty_bed_contact_time.fixed
        assert (
            model.fs.unit.empty_bed_contact_time.value
            == data["empty_bed_contact_time"]["value"] / 60
        )  # min to hour
        assert model.fs.unit.electricity_intensity_parameter.fixed
        assert (
            model.fs.unit.electricity_intensity_parameter.value
            == data["electricity_intensity_parameter"]["value"]
        )
        assert model.fs.unit.activated_carbon_replacement.fixed
        assert (
            model.fs.unit.activated_carbon_replacement.value
            == data["activated_carbon_replacement"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(9.60083, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.00312473, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.083326, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(2.41793, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.49854, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(637.988, rel=1e-5) == value(model.fs.unit.electricity[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


class TestGACZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tss", "nonvolatile_toc", "foo"])

        m.fs.unit = GACZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.empty_bed_contact_time, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.electricity_intensity_constraint, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("gac")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j == "foo":
                assert v.value == 0
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.empty_bed_contact_time.fixed
        assert (
            model.fs.unit.empty_bed_contact_time.value
            == data["empty_bed_contact_time"]["value"] / 60
        )  # min to hour
        assert model.fs.unit.electricity_intensity_parameter.fixed
        assert (
            model.fs.unit.electricity_intensity_parameter.value
            == data["electricity_intensity_parameter"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(9.60183, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.0031244, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.083317, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(0.1041468, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(2.41793, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.49854, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(2.4927e-08, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(638.051, rel=1e-5) == value(model.fs.unit.electricity[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


db = Database()
params = db._get_technology("gac")


class TestGACZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tss", "nonvolatile_toc"])

        m.fs.unit = GACZO(property_package=m.fs.params, database=db)

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("gac", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]


db = Database()
params = db._get_technology("gac")


@pytest.mark.component
@pytest.mark.parametrize("subtype", [k for k in params.keys()])
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = GACZO(
        property_package=m.fs.params, database=m.db, process_subtype=subtype
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.gac, Block)
    assert isinstance(m.fs.costing.gac.contactor_cost_coeff_0, Var)
    assert isinstance(m.fs.costing.gac.contactor_cost_coeff_1, Var)
    assert isinstance(m.fs.costing.gac.contactor_cost_coeff_2, Var)
    assert isinstance(m.fs.costing.gac.contactor_cost_coeff_3, Var)
    assert isinstance(m.fs.costing.gac.adsorbent_unit_cost_coeff, Var)
    assert isinstance(m.fs.costing.gac.adsorbent_unit_cost_exp_coeff, Var)
    assert isinstance(m.fs.costing.gac.other_cost_coeff, Var)
    assert isinstance(m.fs.costing.gac.other_cost_exp, Var)
    assert isinstance(m.fs.costing.gac.bed_mass_max_ref, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]
    assert (
        m.fs.unit1.activated_carbon_demand[0]
        in m.fs.costing._registered_flows["activated_carbon"]
    )
